/* Copyright (c) 2015 Boulder Real Time Technologies, Inc. */
/* All rights reserved */

/* This software module is wholly owned by Boulder Real Time 
	Technologies, Inc. Any use of this software module without
	express written permission from Boulder Real Time Technologies,
	Inc. is prohibited. */

#ifndef __BQ__
#define __BQ__

#include <stddef.h>
#include <cstdarg>
#include <functional>

#include <QtWidgets>

#include "brttutil.h"
#include "coords.h"
#include "stock.h"
#include "BUConfigure.h"
#include "BUMapData.h"
#include "BUEventView.h"

#define BQ_ClientData 	BU_ClientData
#define BQ_ConfigSpec 	BU_ConfigureSpec
#define BQ_CustomOption BU_CustomOption
#define BQ_Offset 	BU_Offset
#define BQ_CONFIGURE_OK 	BU_CONFIGURE_OK
#define BQ_CONFIGURE_ERROR 	BU_CONFIGURE_ERROR
#define BQ_CONFIGURE_NULL_OK 	BU_CONFIGURE_NULL_OK
#define BQ_CONFIGURE_DONT_SET_DEFAULT 	BU_CONFIGURE_DONT_SET_DEFAULT
#define BQ_CONFIGURE_OPTION_SPECIFIED 	BU_CONFIGURE_OPTION_SET
#define BQ_CONFIGURE_BOOLEAN	"boolean"
#define BQ_CONFIGURE_INT	"int"
#define BQ_CONFIGURE_DOUBLE	"double"
#define BQ_CONFIGURE_STRING	"string"
#define BQ_CONFIGURE_PTR	"ptr"
#define BQ_CONFIGURE_CUSTOM	"custom"
#define BQ_CONFIGURE_COLOR	"color"
#define BQ_CONFIGURE_FONT	"font"
#define BQ_CONFIGURE_END	"end"
#define BQ_SetName		BU_SetName
#define BQ_GetName		BU_GetName
#define BQ_GetObject		BU_GetObject
#define BQ_ParseBoolean		BU_ParseBoolean

#define BQ_register_error	bu_register_error
#define BQ_complain		bu_complain
#define BQ_die			bu_die
#define BQ_clear_register	bu_clear_register
#define BQ_elog_string		bu_elog_string

#define BQ_QString2char		toLocal8Bit().data()

#define BQ_ANCHOR_SOUTHWEST	0
#define BQ_ANCHOR_SOUTHEAST	1
#define BQ_ANCHOR_NORTHWEST	2
#define BQ_ANCHOR_NORTHEAST	3

#define BQ_FLOAT_NULL		(2.e30)
#define BQ_DEFAULT_CAP_STYLE	CapButt
#define BQ_DEFAULT_JOIN_STYLE	JoinRound
#define BQ_DEFAULT_LINE_STYLE	LineSolid
#define BQ_X2PS_FACTOR		(10.0)
#define BQ_MINDX		"200"
#define BQ_MINDY		"100"

#define BQ_MODE_LIN		(0)	/* linear plot mode */
#define BQ_MODE_LOG		(1)	/* logarithmic plot mode */
#define BQ_MODE_TIME		(2)	/* time plot mode */

#define BQ_MAP_TRAN_NONE	(0)	/* No map coordinate transformations */
#define BQ_MAP_TRAN_EDP		(1)	/* Equal distance projection map coordinate transformations */
#define BQ_MAP_TRAN_GLOBE	(2)	/* Equal distance projection map coordinate transformations */
#define BQ_MAP_TRAN_MERC	(3)	/* Mercator projection map coordinate transformations (wrapping longitude) */
#define BQ_MAP_TRAN_GEOG	(4)	/* geographic map coordinates (wrapping longitude) */

#define BQ_QEVENT_TYPE_MOUSEPRESS	1
#define BQ_QEVENT_TYPE_MOUSEMOVE	2
#define BQ_QEVENT_TYPE_MOUSERELEASE	3
#define BQ_QEVENT_TYPE_KEYPRESS		4
#define BQ_QEVENT_TYPE_WHEEL		5
#define BQ_QEVENT_TYPE_ENTER		6
#define BQ_QEVENT_TYPE_LEAVE		7
#define BQ_QEVENT_TYPE_RESIZE		8
#define BQ_QEVENT_TYPE_PAINT		9

#define BQ_QEVENT_STATE_INITIAL		1
#define BQ_QEVENT_STATE_PAN		2
#define BQ_QEVENT_STATE_ZOOMIN		3
#define BQ_QEVENT_STATE_ZOOMOUT		4
#define BQ_QEVENT_STATE_RLATLON		5

#define BQ_QEVENT_RETURN_OK		1
#define BQ_QEVENT_RETURN_DONT_PROPAGATE	2

#define BQ_INTERACTION_MODE_NONE	0
#define BQ_INTERACTION_MODE_SHOW	1
#define BQ_INTERACTION_MODE_SHOW_WITHOUT_SIZE_INCREASE	2
#define BQ_INTERACTION_MODE_MOVE	3
#define BQ_INTERACTION_MODE_EVENTSELECT	4

#define BQ_XPIXEL(O,X) ((X)-O->viewport_parent->vars.xleft)*(O->viewport_parent->vars.w - O->viewport_parent->vars.mleft - O->viewport_parent->vars.mright)/(O->viewport_parent->vars.xright-O->viewport_parent->vars.xleft)
#define BQ_XPIXELV(V,X) ((X)-V->vars.xleft)*(V->vars.w - V->vars.mleft - V->vars.mright)/(V->vars.xright-V->vars.xleft)

#define	BQ_MERCY(lat)	(((lat) >= 90.0) ? (1.e5) : (((lat) <= -90.0) ? (-1.e5) \
			: (180.0*log(tan((0.5*M_PI*(lat)/180.0)+0.25*M_PI))/M_PI)))

#define BQ_IMERCLAT(y)	(180.0*(2.0 * (atan(exp((y)*M_PI/180.0)) - 0.25*M_PI))/M_PI)

#define INT(x)  ((x)<0.0?((x)-0.5):((x)+0.5))

struct BQ_Color {
	QString color_name;
	QColor color;
};

struct BQ_Font {
	BQ_Font () {
		memset (this, 0, BQ_Offset (BQ_Font, font_name));
		invalid = 1;
	}

	int size;
	int weight;
	int italic;
	int invalid;
	QString font_name;
	QString family;
	QFont font;
};

struct BQ_XY {
	BQ_XY () {
		x = BQ_FLOAT_NULL;
		y = BQ_FLOAT_NULL;
	}
	BQ_XY (double x, double y) {
		this->x = x;
		this->y = y;
	}

	double x, y;
};

struct BQ_XYL {
	BQ_XYL () {
		x = BQ_FLOAT_NULL;
		y = BQ_FLOAT_NULL;
	}
	BQ_XYL (double x, double y, QString label) {
		this->x = x;
		this->y = y;
		this->label = label;
	}

	double x, y;
	QString label;
};

class BQConfigure : public BUConfigure {
public:
	BQConfigure ();
};

class BQViewportItem;
class BQViewport;
class BQLayer;
class BQLogo;
class BQGraphicsContext;
class BQGlyph;
class BQShape;
class BQGlyphs;
class BQGlyphsGraphicsContext;
class BQAxes;
class BQText;
class BQTextPopup;
class BQTaskbar;
class BQPolyline;
class BQPolypoint;
class BQMap;
class BQNavigator;
class BQMapevents;

class BQViewportItem {
	friend class BQViewport;
	friend class BQLayer;

public:
	BQViewportItem (BQViewport *master, QWidget *widget, char *class_name=NULL, char *item_name=NULL, int register_item=1);
	BQViewportItem (BQViewport *master, QWidget *widget, QString class_name, QString item_name, int register_item=1);
	~BQViewportItem ();
	QString getFullLayerName (QString name);
	void setLayer (QString name);
	void setItemStale();
	void setItemName(QString name);
	void setAntialias(bool antialias);
	void showNames (char *string);
	void removeLayer ();
	bool hasLayer();
	BQLayer *getLayer();
	QString getItemName();
	QString getLayerName();
	BQViewport *getViewportParent();
	void showItem();
	void hideItem();
	void processItem();
	void ignoreItem();
	void moveViewport (BQViewport *viewport);

	int order;
	int item_process;
	int item_display;

protected:
	virtual int processEvent (int type, QEvent *event);
	virtual void drawViewportItem (QPainter *painter) = 0;
	virtual void setVisibility () = 0;
	QPainter *getPainter(QString *layer_name = NULL);

	QWidget *widget;
	BQViewport *viewport_parent;
	QString layer_name;
	QString class_name;
	QString item_name;
	bool antialias;
	static int next_order;
};

class BQLayer : public BQViewportItem {
	friend class BQViewportItem;

public:
	BQLayer (BQViewport *master, QString name);
	void setLayerSize (int width, int height, int *stale);
	void clearLayer();
	void setStale();
	void maskColorTransparent();
	void maskColorTransparent (QColor *mask_color);

	QPixmap *getPixmap();

	int stale;

	static BQLayer *getLayer (QString name);

protected:
	void drawViewportItem (QPainter *painter);
	void removeItem (BQViewportItem *item);
	void addItem (BQViewportItem *item);
	void setVisibility ();

private:
	static QHash<QString, BQLayer *> layers;
	QPixmap *pixmap;
	QColor *mask_color;
	QList<BQViewportItem *> display_list;
};

class BQEvent {
public:
	BQEvent (int type, QEvent *event);

	int type;
	int button;
	int modifiers;
	double delta;
	QPointF pt;
	QPointF gpt;
};

typedef void (BQ_paintCallback) (void *item, int type, QResizeEvent *event, BQ_ClientData client_data);
typedef void (BQ_eventCallback) (void *item, BQEvent *event, BQ_ClientData client_data);

struct BqViewportPaintCallback {
	BqViewportPaintCallback () {
		memset (this, 0, sizeof(BqViewportPaintCallback));
	}

	std::function<BQ_paintCallback> callback_proc;
	void *item;
	BQ_ClientData client_data;
	int stale;
};

struct BqViewportEventCallback {
	BqViewportEventCallback () {
		memset (this, 0, sizeof(BqViewportEventCallback));
	}

	std::function<BQ_eventCallback> callback_proc;
	void *item;
	BQ_ClientData client_data;
};

struct BqViewport {
	BqViewport () {
		memset (this, 0, BQ_Offset (BqViewport, viewport_name));
	}

	int w, h;
	int spc_w, spc_h;
	int last_w, last_h;
	int mleft, mright, mbottom, mtop;
	double opacity;
	int preserve_xscale;
	int preserve_yscale;
	int center_yscale;
	double constant_aspect;
	double xleft, xright, ybottom, ytop;
	double spc_xleft, spc_xright, spc_ybottom, spc_ytop;
	int xoffset, yoffset;
	int xgain_anchor, ygain_anchor;
	double xgain, ygain;
	int xmode, ymode;
	double xmin, xmax, ymin, ymax;
	int map_tran;
	double latr, lonr, range;
	double bblatmin, bblatmax;
	double bblonmin, bblonmax;
	int qevent_state;
	QPointF mouse_start;
	QPointF mouse_start_wc;
	QRubberBand *rubberband;
	BQLogo *logo;
	QResizeEvent *resize_event;
	int mouse_timer_event;
	int wheel_timer_event;
	int default_event_processing;
	QWidget *parent;
	int make_logo;
	BQViewportItem *item_grab;
	QGridLayout *layout;
	QGridLayout *layout_parent;
	int debug;

	QString viewport_name;
	BQ_Color background;
    BQ_Color background_frame;
	QList<BQViewportItem *> display_list;
	QList<BqViewportPaintCallback> paint_callbacks;
	QList<BqViewportEventCallback> event_callbacks;
};

class BQViewport : public QWidget, public BQViewportItem, public BQConfigure {
public:
	BQViewport(QWidget *parent, int make_parent_layout=1, int make_logo=1);

	BqViewport vars;

	static int viewport_next_instance;

	void registerItem (BQViewportItem *item);
	void deregisterItem (BQViewportItem *item);
	void removeItem (BQViewportItem *item);
	void addItemAfter (BQViewportItem *item, BQViewportItem *after_item);
	void moveItemAfter (BQViewportItem *item, BQViewportItem *after_item);
	void moveItemToTop (BQViewportItem *item);
	void showDisplayList ();
	void grabEventStream (BQViewportItem *item);
	void ungrabEventStream ();
	void setPaintCallback (std::function<BQ_paintCallback> paintCallback, void *item, BQ_ClientData client_data);
	void setEventCallback (std::function<BQ_eventCallback> eventCallback, void *item, BQ_ClientData client_data);

	void setScales ();
	int getScales (int *w, int *h,
				double *xstart, double *xend, double *xscale, 
				double *ystart, double *yend, double *yscale);
	int getWcoords (double x, double y, double *xcoord, double *ycoord, int no_map_tran=0, int no_pscale=0);
	QPainterPath *getWcoordsBB (int nside, int *np=NULL, int *sp=NULL);
	int getPcoords (double xcoord, double ycoord, double *x, double *y, int no_map_tran=0, int no_pscale=0);
	int translateAndGain ();
	void bufferDraws (QPainter *painter, int fill, int flush, int w, int h, double fx1, double fy1, double fx2, double fy2);
	int getPixels (char *xstr, char *ystr, double *xpixels, double *ypixels, void *relative=NULL);
	int getPixels (char *str, int *pixels);
	void drawText (QPainter *painter, 
			BQ_Font *font, double x, double y, char *just, double angle, char *text,
			int show_not_symbol=0, int background_width=0, int background_height=0, 
			BQ_Color *text_color=NULL, BQ_Color *background_color=NULL, BQ_Color *background_outline_color=NULL,
			QPainterPath **path=NULL, double *xbb=NULL, double *ybb=NULL, int *wbb=NULL, int *hbb=NULL);
	int polyDraw (QPainter *painter, int npts, double *xpts, double *ypts, int fill, int complete, int ignoretran);
	int drawXTimeAxis (QPainter *painter, BQ_Font *font, 
			BQ_Color *linecolor, int linewidth, int linestyle,
			BQ_Color *fontcolor, char *label_time_format, char *label_date_format);
	void computeMapBB ();
	void setDefaultCursor ();
	void zoom (double xw, double yw, double xfactor, double factor);
	void center (double xw, double yw);
	int processEvent (int type, QEvent *event);
	void setStale ();
	QPainterPath *wrapLongitudes (int npts, double *lons, double *lats, int np, int sp);
	QPainterPath *wrapLongitudes (QPainterPath *path, int np, int sp);

	static int getScales (BqViewport *vp, int *w, int *h,
				double *xstart, double *xend, double *xscale, 
				double *ystart, double *yend, double *yscale);
	static int getWcoords (BqViewport *vp, double x, double y, double *xcoord, double *ycoord, int no_map_tran=0, int no_pscale=0);
	static int getPcoords (BqViewport *vp, double xcoord, double ycoord, double *x, double *y, int no_map_tran=0, int no_pscale=0);
	static int translateAndGain (BqViewport *vp);
	static QPainterPath *wrapLongitudes (BqViewport *vp, int npts, double *lons, double *lats, int np, int sp);
	static QPainterPath *wrapLongitudes (BqViewport *vp, QPainterPath *path, int np, int sp);

protected:
	void paintEvent (QPaintEvent *event);
	void mousePressEvent (QMouseEvent *event);
	void mouseMoveEvent (QMouseEvent *event);
	void mouseReleaseEvent (QMouseEvent *event);
	void wheelEvent (QWheelEvent *event);
	void keyPressEvent (QKeyEvent *event);
	void resizeEvent (QResizeEvent *event);
	void enterEvent (QEvent *event);
	void leaveEvent (QEvent *event);

	void drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
	void paintCallback ();
	void eventCallback (int type, QEvent *event);
	void processevent_pvt (int type, QEvent *event);
	void limitBB (double *x, double *y);
	QRectF boundingRect() const;
	static int setFpoints (int fnpts, int *fpoints_size, QPointF ** fpoints);
	int getXPixel (char *xstr, double *xpixels, void *relative);
	int getYPixel (char *ystr, double *ypixels, void *relative);
	static void parseJustification (char *just, double *jx, double *jy);
};

#define BQ_GRAPHICS_CONTEXT_TYPE_CLEAR_PEN		0
#define BQ_GRAPHICS_CONTEXT_TYPE_COLOR_PEN		1
#define BQ_GRAPHICS_CONTEXT_TYPE_LINESTYLE_PEN		2
#define BQ_GRAPHICS_CONTEXT_TYPE_JOINSTYLE_PEN		3
#define BQ_GRAPHICS_CONTEXT_TYPE_CAPSTYLE_PEN		4
#define BQ_GRAPHICS_CONTEXT_TYPE_CLEAR_BRUSH		5
#define BQ_GRAPHICS_CONTEXT_TYPE_COLOR_BRUSH		6
#define BQ_GRAPHICS_CONTEXT_TYPE_LINEWIDTH		7
#define BQ_GRAPHICS_CONTEXT_TYPE_LINEWIDTHF		8
#define BQ_GRAPHICS_CONTEXT_TYPE_OPACITY		9
#define BQ_GRAPHICS_CONTEXT_TYPE_LABEL			10
#define BQ_GRAPHICS_CONTEXT_TYPE_FONT_LABEL		11
#define BQ_GRAPHICS_CONTEXT_TYPE_COLOR_LABEL		12
#define BQ_GRAPHICS_CONTEXT_TYPE_SIZE_LABEL		13
#define BQ_GRAPHICS_CONTEXT_TYPE_SYMBOL			14
#define BQ_GRAPHICS_CONTEXT_TYPE_SIZE			15

#define BQ_GRAPHICS_CONTEXT_CODING_TYPE_NONE		0
#define BQ_GRAPHICS_CONTEXT_CODING_TYPE_SCALAR		1
#define BQ_GRAPHICS_CONTEXT_CODING_TYPE_COLOR_HSL	2

#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_DEFAULT		(0)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_POINT		(1)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_CROSS		(2)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_X		(3)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_TRIANGLE	(4)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_SQUARE		(5)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_CIRCLE		(6)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_STAR		(7)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_DIAMOND		(8)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_BEACHBALL	(9)
#define BQ_GRAPHICS_CONTEXT_SYMBOL_TYPE_SHAPE		(10)

struct BqGraphicsContextCoding {
	BqGraphicsContextCoding () {
		memset (this, 0, sizeof (BqGraphicsContextCoding));
	}

	double in_0;
	double out_0;
	double out_scale;
	double out_min;
	double out_max;
};

struct BqGraphicsContextItemParams {
	BqGraphicsContextItemParams () {
		memset (this, 0, BQ_Offset (BqGraphicsContextItemParams, name));
	}
	~BqGraphicsContextItemParams () {
		QColor *color;
		QFont *font;
		if (free_flag) return;
		if (coding_type == BQ_GRAPHICS_CONTEXT_CODING_TYPE_COLOR_HSL) {
			color = (QColor *) coding_out;
			if (color) delete color;
		}
		switch (type) {
		case BQ_GRAPHICS_CONTEXT_TYPE_COLOR_PEN:
		case BQ_GRAPHICS_CONTEXT_TYPE_COLOR_BRUSH:
		case BQ_GRAPHICS_CONTEXT_TYPE_COLOR_LABEL:
			color = (QColor *) value;
			if (color) delete color;
			break;
		default:
			break;
		}
	}

	int type;
	void *value;
	int coding_type;
	void *coding_out;
	BqGraphicsContextCoding *coding;
	int free_flag;
	QString name;
};

struct BqGraphicsContextItem {
	BqGraphicsContextItem () {
		memset (this, 0, sizeof(BqGraphicsContextItem));
		coding_input = BQ_FLOAT_NULL;
	}
	~BqGraphicsContextItem () {
		if (item && item->free_flag) delete item;
	}

	BqGraphicsContextItemParams *item;
	float coding_input;
};

typedef QList<BqGraphicsContextItem> BqGraphicsContext;

class BQGraphicsContext {
public:
	BQGraphicsContext ();
	BQGraphicsContext (BQGraphicsContext *source);
	void copy (BQGraphicsContext *source);
	BqGraphicsContextItem *setGcItem (QString name, int type, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, QColor param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, QFont *param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, Qt::PenJoinStyle style, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, long param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, double param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, QString param, int replace=0);
	BqGraphicsContextItem *setGcItem (QString name, int type, char *param, int replace=0);
	int parseColorCoding (QString string, QColor **constant_color, BqGraphicsContextCoding **color_coding);
	BqGraphicsContext *getGc (QString name);
	BqGraphicsContextItem *getGcItem (QString name);
	void setGcItemCoding (QString name, int coding_type, BqGraphicsContextCoding coding);
	void setGcItemCoding (QString name, int coding_type, BqGraphicsContextCoding *coding = NULL);
	void drawGcItem (QString name, QPainter *painter);
	void drawGc (QString name, QPainter *painter);
	void drawPainterPath (QString name, QPainter *painter, QPainterPath &path);

	static BqGraphicsContext *copy (BqGraphicsContext *gc);
	static void getValue (const BqGraphicsContextItem *item, void **value);
	static void getValue (const BqGraphicsContextItem *item, int *value);
	static void getValue (const BqGraphicsContextItem *item, long *value);
	static void getValue (const BqGraphicsContextItem *item, double *value);
	static void setGcCoding (BqGraphicsContextCoding *coding, double in1, double out1, double in2, double out2, double outmin=1.0, double outmax=-1.0);
	static void setGcItemCoding (BqGraphicsContextItem *item, int coding_type, BqGraphicsContextCoding coding);
	static void setGcItemCoding (BqGraphicsContextItem *item, int coding_type, BqGraphicsContextCoding *coding = NULL);
	static void drawGcItem (const BqGraphicsContextItem *item, QPen *pen, QBrush *brush);
	static void drawGcItem (BqGraphicsContextItem *item, QPainter *painter);
	static void makeGcPenBrush (BqGraphicsContext *gc, QPen *pen, QBrush *brush);
	static void drawGc (BqGraphicsContext *gc, QPainter *painter);
	static void drawGcPainter (BqGraphicsContext *gc, QPainter *painter);
	static void drawGcPainter (BqGraphicsContext *gc, QPen *pen, QBrush *brush, QPainter *painter);
	static void drawPainterPath (BqGraphicsContext *gc, QPainter *painter, QPainterPath &path);
	static BqGraphicsContextItem *getSymbol (int symbol);
	static void initContextItem (BqGraphicsContextItem *item);

	QHash<QString, BqGraphicsContext *> mapgc;
	QHash<QString, BqGraphicsContextItem *> mapgcitem;

private:
	BqGraphicsContext *get_gc (QString name);
	BqGraphicsContextItem *get_item (QString name, int set = 1, int replace=0);
	int parseCodings (QString string, int ncodings, BqGraphicsContextCoding *color_coding);
	static void makeStaticSymbols();
	static void setStaticSymbol(QString name, int symbol_type);

	static QHash<QString, BqGraphicsContextItem *> *symbols;
};

struct BqShape {
	BqShape () {
		memset (this, 0, BQ_Offset (BqShape, paths));
	}

	QPainterPath *outline;
	QList<QPainterPath *> paths;
	QRectF bounding_rect;
};

class BQShape {
public:
	BQShape();

	static BqShape *getPoint();
	static BqShape *getCross();
	static BqShape *getX();
	static BqShape *getCircle();
	static BqShape *getSquare();
	static BqShape *getTriangle();
	static BqShape *getDiamond();
	static BqShape *getStar();
	static BqShape *getBeachball (double m[6]);
};

#define BQ_GLYPH_GRAPHICS_CONTEXT_SYMBOL	(-5)
#define BQ_GLYPH_GRAPHICS_CONTEXT_SIZE		(-4)

typedef int (BQ_glyphsGraphicsContextCallback) (BQGlyph *glyph, QString definitions_name, QString coding_name, double *value, BQ_ClientData client_data);

struct BqGlyphsGcDefinition {
	BqGlyphsGcDefinition () {
		memset (this, 0, BQ_Offset (BqGlyphsGcDefinition, coding));
	}

	BqGraphicsContextItem *gcitem;
	QString coding;
	QString value;
};

struct BqGlyphsCodingSlot {
	BqGlyphsCodingSlot () {
		memset (this, 0, sizeof (BqGlyphsCodingSlot));
	}

	int gc;
	int gci;
};

struct BqGlyphsDefinition {
	QString name;
	BqGraphicsContextItem symbol;
	BqGlyphsGcDefinition size;
	QList<BqGlyphsGcDefinition> fill_color_items;
	QList<BqGlyphsGcDefinition> outline_linewidth_items;
	QList<BqGlyphsGcDefinition> outline_color_items;
	QList<BqGlyphsGcDefinition> opacity_items;
	QList<BqGraphicsContext *> glyph_gcs;
	QHash<QString, int> coding_defs;
	QList<QList<BqGlyphsCodingSlot>> coding_slots;
};

class BQGlyphsGraphicsContext : public BQGraphicsContext, public BQConfigure {
	friend class BQGlyphs;
	friend class BQGlyph;

public:
	BQGlyphsGraphicsContext();
	~BQGlyphsGraphicsContext();

	int getGcIndex (QString gcname);
	int getCodingIndex (int gcindex, QString codingname);
	void setCoding (int gcindex, int codingindex, double coding_input, BQGlyph *glyph);
	void getCoding (int gcindex, int codingindex, double *coding_input, BQGlyph *glyph);
	void setCodingCallback (std::function<BQ_glyphsGraphicsContextCallback> codingCallback, BQ_ClientData client_data);

	QHash<QString, BqGlyphsGcDefinition> gcitem_definitions;
	QHash<QString, BqGlyphsDefinition> gc_definitions;
	QList<BqGlyphsDefinition> gc_definitions_list;
	QHash<QString, int> gc_definitions_index;
	QList<QString> coding_names;

protected:
	void updateCoding (BQGlyph *glyph);
	BQ_ClientData codingClientData;
	std::function<BQ_glyphsGraphicsContextCallback> codingCallback;
};

struct BqPath {
	BqPath () {
		memset (this, 0, BQ_Offset (BqPath, pen));
	}

	BqGraphicsContext *gc;
	QPainterPath *path;
	QPen pen;
	QBrush brush;
};

struct BqGlyph {
	BqGlyph () {
		memset (this, 0, BQ_Offset (BqGlyph, gcs));
	}

	int index;
	void *shape;
	double xw, yw;
	double xp, yp;
	double opacity;
	int visible;
	int skip;
	int visible_label;
	char *label;
	int qevent_state;
	int freegc;
	BQGlyphs *glyphs;
	int gcindex;

	BQ_ClientData clientData;
	int clientData_type;
	QList<BqGraphicsContext *> gcs;
	QPainterPath shape_path;
	QList<BqPath> paths;
	QList<QList<QPainterPath>> plot_paths;
	BqGraphicsContextItem symbol;
	BqGraphicsContextItem size;
};

class BQGlyph : public BQViewportItem {
	friend class BQGlyphs;
	friend class BQPolypoint;
	friend class BQGlyphsGraphicsContext;

public:
	BQGlyph(BQViewport *master, BQGlyphs *glyphs);
	~BQGlyph();
	void setGlyph (int symbol, double *params, QList<BqGraphicsContext *> gcs);
	void setGlyph (int symbol, double *params, BqGraphicsContext *gc);
	void setGlyph (QString gc_definitions_name, double *params);
	void setGlyph (int symbol, QString gc_definitions_name, double *params);
	void setPosition (double xw, double yw);
	void getPosition (double *xw, double *yw);
	int getIndex ();
	int getQEventState ();
	void setSize (double size);
	void setCodingValue (QString name, double coding_value);
	double getCodingValue (QString name);
	bool contains (QMouseEvent *event);
	void setClientData(BQ_ClientData client_data, int client_data_type=0);
	void getClientData(BQ_ClientData *client_data, int *client_data_type=NULL);
	BqGraphicsContextItem getGraphicsContextItem (int gcindex, int type);
	void showPaths();

	mutable BqGlyph vars;

protected:
	int processEvent (int type, QEvent *event);
	void drawViewportItem (QPainter *painter) {};
	void setVisibility () {};
	void setGlyph (double *params);
	void draw(QPainter *painter);
	void drawSelected(QPainter *painter);
	void drawUnselected(QPainter *painter);
	void setCodingInput (int igc, int igci, double coding);
	void getCodingInput (int igc, int igci, double *coding);
	void setCodingInputs (int def_index, int coding_index, double coding);

private:
	void clearPaths ();
	int setTransformations (QList<BQ_XY> *label_positions = NULL) const;
	int processEventMove (int type, QEvent *event);
	int processEventShow (int type, QEvent *event);
	void setGlyph (int gc_definitions_index, double *params);
	void setGlyph (int symbol, int gc_definitions_index, double *params);
};

typedef int (BQ_glyphCallback) (BQGlyph *glyph, BQEvent *event, BQ_ClientData client_data);

struct BqGlyphCallback {
	BqGlyphCallback () {
		memset (this, 0, sizeof(BqGlyphCallback));
	}

	std::function<BQ_glyphCallback> callback_proc;
	BQ_ClientData client_data;
};

struct BqGlyphs {
	BqGlyphs () {
		memset (this, 0, BQ_Offset (BqGlyphs, glyphs));
	}

	BqShape *shape_default;
	double sizep_default;
	BqGraphicsContext *gc_label;
	BQGlyphsGraphicsContext *ggc;
	int selection_index;
	int interaction_mode;

	QList<BQGlyph *> glyphs;
	QList<BqGlyphCallback> glyphCallbacks;
};

class BQGlyphs : public BQViewportItem {
	friend class BQGlyph;

public:
	BQGlyphs(BQViewport *master, BQGlyphsGraphicsContext *ggc = NULL);
	~BQGlyphs();
	void setShapeDefault (int symbol);
	void setSizeDefault (double sizep);
	void setLabelGC (BqGraphicsContext *gc);
	BQGlyph *selectedGlyph();
	QList<BQGlyph *> items (QMouseEvent *mevent);
	void setInteractionMode (int interaction_mode);
	void setGlyphCallback (std::function<BQ_glyphCallback> glyphCallback, BQ_ClientData client_data);
	void setGlyphCallbacks (QList<BqGlyphCallback> callbacks);
	void setGlyphs();
	void updateCoding();
	void setClientData (BQ_ClientData client_data);
	BQ_ClientData getClientData ();

	BqGlyphs vars;

protected:
	void drawViewportItem (QPainter *painter);
	int processEvent (int type, QEvent *event);
	void addGlyph (BQGlyph *glyph);

private:
	void setVisibility () {};
	BQGlyph *findClosest (QEvent *event);

	BQ_ClientData clientData;
};

struct BqLogo {
	BqLogo () {
		memset (this, 0, sizeof (BqLogo));
	}

	double opacity;
	QImage *logo_image;
};

class BQLogo : public BQViewportItem, public BQConfigure {
public:
	BQLogo(BQViewport *master);
	void raiseLogo();

	BqLogo vars;

protected:
	void drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
	void draw (QPainter *painter);
};

struct BqAxes {
	BqAxes () {
		memset (this, 0, BQ_Offset (BqAxes, foreground_labels));
	}

	int xvisible_tics;
	int yvisible_tics;
	char *x_label;
	char *y_label;
	char *x_format;
	char *y_format;
	char *x_time_format;
	char *x_date_format;
	char *axis_style;
	char *tic_style;
	char tic_style_west;
	char tic_style_north;
	char tic_style_east;
	char tic_style_south;
	double xincr_num;
	double xincr_sml;
	double yincr_num;
	double yincr_sml;
	int boxwidth;
	int xtimewidth;
	int xwidth_big;
	int xwidth_sml;
	double xticlength_sml;
	double xticlength_big;
	int ywidth_big;
	int ywidth_sml;
	double yticlength_sml;
	double yticlength_big;
	int mindx;
	int mindy;
	int style;
	int xvisible_grids;
	int yvisible_grids;
	int xwidth_big_grids;
	int xwidth_sml_grids;
	int ywidth_big_grids;
	int ywidth_sml_grids;

	BQ_Color foreground_labels;
	BQ_Color foreground_numbers;
	BQ_Color foreground_axes;
	BQ_Color x_time_axes;
	BQ_Font font_labels;
	BQ_Font font_numbers;
	BQ_Font font_supers;
	BQ_Color color_grids_x;
	BQ_Color color_grids_x_small;
	BQ_Color color_grids_y;
	BQ_Color color_grids_y_small;
};

class BQAxes : public BQViewportItem, public BQConfigure {
public:
	BQAxes(BQViewport *master);
	int processEvent (int type, QEvent *event);

	BqAxes vars;

protected:
	void drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
	void chooseScale (double range, int min_ticks, int max_ticks, 
			double *divisions, int ndivisions, double *major, double *minor, int *power);
	char *chooseTimeFormat ( double major );
	void labelXTic (QPainter *painter, char *string, double x, double fy);
	void drawXTimeAxis (QPainter *painter, double t0, double t1);
	void drawXTicLabels (QPainter *painter, int w, double xstart, double xscale);
	void drawYTicLabels (QPainter *painter, int h, double ystart, double yscale);
	void drawBox (int w, int h, QPainter *painter);
	void drawTics (int w, int h, QPainter *painter);
	void drawXLabel (QPainter *painter);
	void drawYLabel (QPainter *painter);
	void drawXaxisStuff (QPainter *painter, 
		BQ_Color * color, int linewidth, int linestyle, double incr, double fy1, double fy2,
		BQ_Color * color_small, int linewidth_small, int linestyle_small, double incr_small,
		double fy1_small, double fy2_small, double mindx);
	void drawYaxisStuff (QPainter *painter, 
		BQ_Color *color, int linewidth, int linestyle, double incr, double fx1, double fx2,
		BQ_Color *color_small, int linewidth_small, int linestyle_small, double incr_small,
		double fx1_small, double fx2_small, double mindy);
	void draw (QPainter *painter);
}; 

typedef int (BQ_textCallback) (BQText *text, int type, QEvent *event, BQ_ClientData client_data);

struct BqTextCallback {
	BqTextCallback () {
		memset (this, 0, sizeof(BqTextCallback));
	}

	std::function<BQ_textCallback> callback_proc;
	BQ_ClientData client_data;
};

struct BqText {
	BqText () {
		memset (this, 0, BQ_Offset (BqText, foreground));
	}

	double opacity;
	int select;
	char *text;
	double fxbb, fybb;
	int wbb, hbb;
	double x, y;
	double angle;
	char *justification;
	void *relative;
	char *xpos;
	char *ypos;
	int width, height;
	int show_not_symbol;
	QPainterPath *path;
	BQ_ClientData textCallbackClientData;

	BQ_Color foreground;
	BQ_Color outline;
	BQ_Color background;
	BQ_Font font;
	QList<BqTextCallback> textCallbacks;
};

struct BqTextSpecialSymbol {
	BqTextSpecialSymbol () {
		memset (this, 0, sizeof (BqTextSpecialSymbol));
	}

	char *name;
	int size;
	int width;
	int height;
	QPixmap *pixmap;
};

class BQText : public BQViewportItem, public BQConfigure {
	friend class BQTaskbar;

public:
	BQText(BQViewport *master, QString textstring, QString x, QString y, int register_item = 1);
	~BQText();
	int processEvent (int type, QEvent *event);
	void getBB (double *x, double *y, int *w, int *h);
	static QPixmap *specialSymbol (char *name, int size);
	void setTextCallback (std::function<BQ_textCallback> textCallback, BQ_ClientData client_data);

	BqText vars;

protected:
	void drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init(char *textstring, char *x, char *y);
	void init(QString textstring, QString x, QString y);
	bool contains (QMouseEvent *event);
	void draw (QPainter *painter);
	static BqTextSpecialSymbol *makeSpecialSymbol (char *name, int size);
}; 

class BQTextPopupWidget : public QWidget {
public:
	BQTextPopupWidget(BQTextPopup *popup);
	~BQTextPopupWidget();


protected:
	void paintEvent (QPaintEvent *event);

	BQTextPopup *popup;
};

struct BqTextpopup {
	BqTextpopup () {
		memset (this, 0, BQ_Offset (BqTextpopup, color_background));
	}

	double opacity;
	double timeout;
	QTimer *timer;
	BQTextPopupWidget *popup_widget;
	QPixmap *pixmap;

	BQ_Color color_background;
	QString text;
	QPoint position;
	QString justification;
};

class BQTextPopup : public BQViewportItem, public BQConfigure {
public:
	BQTextPopup(BQViewport *master);
	~BQTextPopup();
	void raiseTextPopup();
	void setText (QString text);
	void setTimeout (double timeout);
	void setJustification (QString justification);
	void setPosition (QPoint pos);
	void stopTimer ();
	void deletePopup ();
	QPoint justify();
	static QPixmap *makePixmap(QString text, QColor *background);

	BqTextpopup vars;

protected:
	void drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
};

class BQPolylinePath {
public:
	BQPolylinePath (QPainterPath *path, int np=0, int sp=0) {
		this->path = path;
		this->np = np;
		this->sp = sp;
	}
	~BQPolylinePath () {
		if (this->path) delete this->path;
	}

	QPainterPath *path;
	int np;
	int sp;
};

struct BqPolyline {
	BqPolyline () {
		memset (this, 0, BQ_Offset (BqPolyline, fill));
	}
	~BqPolyline () {
		if (xpts) free (xpts);
		if (ypts) free (ypts);
	}

	int linewidth;
	int linestyle;
	double xmin, xmax, ymin, ymax;
	int npts;
	double *xpts;
	double *ypts;
	int xpts_size;
	int ypts_size;
	char *vechandle;
	char *thandle;
	BQPolylinePath *plpath;

	BQ_Color fill;
	BQ_Color outline;
	BQ_Color fill_black;
	BqGraphicsContext gc;
};

class BQPolyline : public BQViewportItem, public BQConfigure {
public:
	BQPolyline(BQViewport *master);
	void setGc (BqGraphicsContext gc);

	BqPolyline vars;

protected:
	void drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init();
	void draw (QPainter *painter);
};

struct BqPolypoint {
	BqPolypoint () {
		memset (this, 0, BQ_Offset (BqPolypoint, glyphCallbacks));
	}
	~BqPolypoint () {
		if (points) delete points;
	}

	int symbol_default;
	double size_default;
	int showlabels;
	BQGraphicsContext *gc_default;
	BqGraphicsContextItem *color_fill_default;
	BqGraphicsContextItem *color_fill_node_default;
	BqGraphicsContextItem *color_outline_default;
	BqGraphicsContextItem *color_text_default;
	BqGraphicsContextItem *font_default;
	BqGraphicsContextItem *linestyle_default;
	BqGraphicsContextItem *joinstyle_default;
	BqGraphicsContextItem *capstyle_default;
	BqGraphicsContextItem *linewidth_default;
	BqGraphicsContextItem *linewidth_node_default;
	BqGraphicsContextItem *opacity_default;
	double linewidth_value_default;
	double linewidth_node_value_default;
	double opacity_value_default;
	int interaction_mode;
	char *vechandle;
	BQGlyphs *points;

	QList<BqGlyphCallback> glyphCallbacks;
	QString layer_name;
	QString color_fill_name_default;
	QString color_fill_node_name_default;
	QString color_outline_name_default;
	QString color_text_name_default;
	QString font_name_default;
	QString linestyle_name_default;
	QString joinstyle_name_default;
	QString capstyle_name_default;
};

class BQPolypoint : public BQViewportItem, public BQConfigure {
public:
	BQPolypoint(BQViewport *master);
	void setLayer (QString layer_name);
	void setGlyphs ();
	void setGlyphCallback (std::function<BQ_glyphCallback> glyphCallback, BQ_ClientData client_data);

	BqPolypoint vars;

protected:
	void drawViewportItem (QPainter *painter) {};
	void setVisibility () {};

private:
	void init();
};

struct BqTaskbarButtonStateParams;

struct BqTaskbarButtonParams {
	BqTaskbarButtonParams () {
		memset (this, 0, BQ_Offset (BqTaskbarButtonParams, description));
	}

	BQTaskbar *taskbar;
	int width, height;
	BQText *bq_text;
	int index;
	int current;

	QString description;
	QList<BqTaskbarButtonStateParams> states;
};

struct BqTaskbarButtonStateParams {
	BqTaskbarButtonStateParams () {
		memset (this, 0, BQ_Offset (BqTaskbarButtonStateParams, name));
	}

	double opacity;
	int show_not_symbol;
	int order;

	QString name;
	QString label;
	QString description;
	QList<QString> action;
	BQ_Color fill_color;
	BQ_Color background_color;
};

struct BqTaskbar {
	BqTaskbar () {
		memset (this, 0, BQ_Offset (BqTaskbar, background));
	}

	int selection_index;
	BQTextPopup *popup;

	BQ_Color background;
	QString buttons_definitions_string;
	QHash<QString, BqTaskbarButtonParams *> buttons_definitions;
	QString buttons_string;
	QList<BqTaskbarButtonParams *> buttons_list;
};

class BQTaskbar : public BQViewportItem, public BQConfigure {
public:
	BQTaskbar(BQViewport *master);
	~BQTaskbar();

	int setState (QString name_button, QString name_state, 
			BqTaskbarButtonParams **params=NULL, BqTaskbarButtonStateParams **state=NULL);

	BqTaskbar vars;

protected:
	void drawViewportItem (QPainter *painter);
	int processEvent (int type, QEvent *event);
	void setVisibility () {};

private:
	void init();
	QList<BQText *> items (QMouseEvent *mevent);
	BQText *findClosest (QEvent *event);
};

class BUMapTransform {
public:
	BUMapTransform (double latr, double lonr);
	void edpDelxy2latlon (double delx, double dely, double *lat, double *lon);
	void edpLatlon2delxy (double lat, double lon, double *delx, double *dely);
	int globeDelxy2latlon (double delx, double dely, double *lat, double *lon);
	int globeLatlon2delxy (double lat, double lon, double *delx, double *dely);
	double latr, lonr, clat, slat;
};

class BQScaledata {
public:
	static const int disposition_noInput = -1;
	static const int disposition_notUsed = 0;
	static const int disposition_moveTo = 1;
	static const int disposition_lineTo = 2;

	BqViewport *vp;
	int map_tran;
	int xmode;
	int ymode;
	int w, h;
	double xstart, xend, xscale;
	double ystart, yend, yscale;
	BUMapTransform *tran;
	double xp_last, yp_last;
	double xpt, ypt;
	double xptp, yptp;
	double xinput, yinput;
	double xpts_last, ypts_last;
	int start, ndelta, idelta;
	double delx, dely;

	BQScaledata (BqViewport *viewport, int ignoretran=0);
	~BQScaledata ();
	int pushPoint (double xpt, double ypt);
	int popPoint (double *xscaled, double *yscaled);
};

class BQPolyDraw {
public:
	BQScaledata *scale;
	int w, h;
	double x_last, y_last;
	int ret_last, zone_last;
	int first;

	BQPolyDraw (BqViewport *viewport, int ignoretran=0);
	~BQPolyDraw ();
	QPainterPath *addPoints (QPainterPath *path, int npts, double *xpts, double *ypts, int start=0, int subtract=0);
	QPainterPath *addPoints (QPainterPath *path, QPainterPath *inpath, int start=0, int subtract=0);
};

class BQMapData : public BUMapData {
public:
	BQMapData (char *fname = NULL, int nthreads = 1);
	BQMapData (int nthreads = 1);
	int setResolution (int resolution);
	int setResolution (double degrees_per_pixel);
	void drawVector(QPainter &painter, BqViewport *viewport, 
				unsigned int levels, BQ_Color *outline, 
				int *linewidth, int *resolution, BQ_Color &water, BQ_Color &land);
	QImage *drawImage(int width, int height, int tran, double xleft, double xright, 
				double ybottom, double ytop, double lonr, double latr, double gamma=0.8, double lightmin=0.0);
	unsigned char *getImage (int ibin);

	int resolution;
	int nthreads;
	unsigned char *image_buf;
	size_t image_buf_size;
	QImage *image;
	int width, height;
	double lightmin;
	double gamma;
	double xmin, xmax, ymin, ymax, dx, dy;
	int tran;
	double latr, lonr;

private:
	Pmtcomp *mtc;
	int isInside(double lat, double lon, int npts, double *latpts, double *lonpts);
};

struct BqNavigator {
	BqNavigator () {
		memset (this, 0, BQ_Offset (BqNavigator, fill_land));
	}

	BQViewport *viewport;
	BQMap *parent_map;
	BQMap *map;
	BQPolyline *map_outline;
	int qevent_state;
	double event_xstart, event_ystart;
	int show;

	BQ_Color fill_land;
	BQ_Color fill_water;
};

class BQNavigator : public BQConfigure {
public:
	BQNavigator (BQMap *parent_map);
	void refreshVisibility();
	void draw ();

	BqNavigator vars;

protected:
	void drawViewportItem (QPainter *painter) {};
	void setVisibility () {};

private:
	int init (BQMap *parent_map);
};

#define	BQ_MAP_MAXLEVELS		(10)
#define	BQ_MAP_MAXLEVELS_COASTS		(8)
#define	BQ_MAP_MAXLEVELS_RIVERS		(10)
#define	BQ_MAP_MAXLEVELS_POLITICAL	(4)


struct BqMapLayer {
	BqMapLayer () {
		memset (this, 0, BQ_Offset (BqMapLayer, fill_color));
	}

	int type;
	double opacity;
	int antialias;
	char *mapdata_source;
	BQMapData *mapdata;
	int nodata;
	unsigned int levels;
	int linewidth[BQ_MAP_MAXLEVELS];
	int resolution[BQ_MAP_MAXLEVELS];
	BQMap *map;

	BQ_Color fill_color;
	BQ_Color background_color;
	BQ_Color transparent_color;
	BQ_Color outline[BQ_MAP_MAXLEVELS];
};

class BQMapLayer : public BQViewportItem, public BQConfigure {
	friend class BQMap;

public:
	BQMapLayer(BQViewport *master, BQMap *map);
	~BQMapLayer();

	BqMapLayer vars;

protected:
	void drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init (BQMap *map);
	void draw (QPainter *painter, char *name);
};

struct BqMap {
	BqMap () {
		memset (this, 0, BQ_Offset (BqMap, fill_land));
	}

	int nthreads;
	int cache_maxsize;
	BQTaskbar *taskbar;
	BQNavigator *navigator;
	BQAxes *grids;
	QWidget *parent;
	int make_navigator;
	int vector_only;

	BQ_Color fill_land;
	BQ_Color fill_water;
	BQ_Color fill_black;
	QHash<QString, QString > maplayer_definitions_string;
	QHash<QString, BQMapLayer * > maplayer_definitions;
	QList<QString> maplayers;
	QHash<QString, QString > maplayers_templates;
	QString taskbar_buttons_definitions_string;
	QString taskbar_buttons_string;
};

#define	BQ_MAP_TYPE_FILL_LAND		(0)
#define	BQ_MAP_TYPE_FILL_WATER		(1)
#define	BQ_MAP_TYPE_LINE_COASTS		(2)
#define	BQ_MAP_TYPE_LINE_RIVERS		(3)
#define	BQ_MAP_TYPE_LINE_POLITICAL	(4)
#define	BQ_MAP_TYPE_GRID		(5)
#define	BQ_MAP_TYPE_VECTOR		(6)
#define	BQ_MAP_TYPE_VECTOR_FILL		(7)

class BQMap : public BQViewportItem, public BQConfigure {
public:
	BQMap(BQViewport *master, int make_navigator=1, int make_grids=1);
	BQMap(QWidget *parent, int make_navigator=1, int make_grids=1);

	BQTaskbar *getTaskbar();
	void clearMaplayerDefinitions();
	void setMaplayersStale ();
	void setDefaults();
	int makeMaplayer (char *name);

	static void solarPosition (double epoch, double *latitude, double *longitude);

	BqMap vars;

protected:
	void drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init (int make_navigator, int make_grids);
};

struct BqDrawEvent {
	BqDrawEvent () {
		memset (this, 0, BQ_Offset (BqDrawEvent, layer));
	}

	BQViewport *viewport;
	BQMap *parent_map;
	BQMap *map;
	BQGlyphsGraphicsContext *ggc;
	BQMapevents *mapevents;
	BQGlyphs *origins_glyphs;
	BUEVOrigin *selected_origin;
	BUEVOrigin *preferred_origin;

	QString layer;
	QString origins_def;
	QString origins_bb_def;
	QString origins_pref_def;
	QString origins_pref_bb_def;
	QList<BQPolyline *> error_ellipse_polylines;
	QHash<int, BQPolyline *> error_ellipse_hash;
};

class BQDrawEvent : public BQConfigure {
public:
	BQDrawEvent (BQViewport *master, BQMap *parent_map,
				BQGlyphsGraphicsContext *ggc, BQMapevents *mapevents, QString layer, QHash<QString, QString> gc_defs);
	void makeOrigins (BUEVEvent *event);
	void hideOrigins ();
	void showOrigins ();
	void showErrorEllipses (int index);

	static int origins_glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	static QList<BQ_XY> *makeErrorEllipse (double lat, double lon, double semimajor, double semiminor, double azimuth, int n=0);

	BqDrawEvent vars;
};

#define BQ_GLYPHS_TYPE_EVENTS	1
#define BQ_GLYPHS_TYPE_ORIGINS	2

struct BqStation {
	BqStation () {
		memset (this, 0, BQ_Offset (BqStation, sta));
	}

	double lat;
	double lon;
	double elev;
	double distance;

	QString sta;
	BqGraphicsContext gc;
};

typedef void (BQ_glyphMapeventsCallback) (BQGlyph *glyph, BQEvent *event, int glyph_type, int show_mode, BQ_ClientData client_data);

struct BqGlyphMapeventsCallback {
	BqGlyphMapeventsCallback () {
		memset (this, 0, sizeof(BqGlyphMapeventsCallback));
	}

	std::function<BQ_glyphMapeventsCallback> callback_proc;
	BQ_ClientData client_data;
};

struct BqMapevents {
	BqMapevents () {
		memset (this, 0, BQ_Offset (BqMapevents, events_definitions));
	}
	~BqMapevents () {
		if (ev && events_tbl) ev->freeEvents(events_tbl);
		if (map_allocated) delete map_allocated;
		if (events_glyphs) delete events_glyphs;
		if (event_popup) delete event_popup;
		if (arrival_popup) delete arrival_popup;
		if (origins_glyphs) delete origins_glyphs;
		if (arrivals_glyphs) delete arrivals_glyphs;
		if (detections_glyphs) delete detections_glyphs;
		if (ggc) delete ggc;
		if (taskbar) delete taskbar;
	}

	BUEventView *ev;
	Tbl *events_tbl;
	Tbl *events_detections_tbl;
	BQMap *map_allocated;
	BQMap *map;
	BQGlyphsGraphicsContext *ggc;
	BQDrawEvent *event_zoomed;
	BQViewport *event_zoomed_vp;
	BQTaskbar *taskbar;
	BQTextPopup *event_popup;
	BQTextPopup *arrival_popup;
	BQGlyphs *events_glyphs;
	BQGlyphs *events_detections_glyphs;
	int events_stale;
	BUEVEvent *selected_event;
	int selected_event_free;
	BQGlyphs *origins_glyphs;
	int origins_stale;
	BUEVOrigin *selected_origin;
	int selected_origin_free;
	BUEVOrigin *preferred_origin;
	BQGlyphs *arrivals_glyphs;
	BQGlyphs *detections_glyphs;
	int show_events_detections;
	int auto_center;
	int auto_show_event;
	double update_interval;
	QTimer *update_timer;
	double max_age_detections;
	int events_interaction_mode;
	double last_time;
	double age_ref;
	int show_mode;

	QString events_definitions[2];
	QList<BQPolyline *> origin2sta_polylines;
	QString taskbar_buttons_definitions_string;
	QString taskbar_buttons_string;
	QString glyphs_gcitem_definitions_string;
	QString glyphs_gc_definitions_string;
	QList<BqGlyphMapeventsCallback> glyphCallbacks;
};

#define BQ_MAPEVENTS_GLYPH_TYPE_EVENT		0
#define BQ_MAPEVENTS_GLYPH_TYPE_ORIGIN		1
#define BQ_MAPEVENTS_GLYPH_TYPE_ASSOCIATION	2
#define BQ_MAPEVENTS_GLYPH_TYPE_DETECTION	3

#define BQ_MAPEVENTS_SHOW_MODE_EVENTS		1
#define BQ_MAPEVENTS_SHOW_MODE_ORIGINS		2

class BQMapevents : public BQViewportItem, public BQConfigure {
public:
	BQMapevents(BQMap *map);
	BQMapevents(BQViewport *master, int make_navigator=1, int make_grids=1);
	BQMapevents(QWidget *parent, int make_navigator=1, int make_grids=1);
	BQMap *getMap();
	BQTaskbar *getTaskbar();
	int setDefaults(int nort=0);
	void setEventView(BUEventView *ev);
	void refreshEventView();
	void updateDisplays (QString which);
	void makeEvents ();
	void makeOrigins (BUEVEvent *event);
	void showEvent (BUEVEvent *event, int event_free=0);
	void showOrigin (BUEVOrigin *origin, int origin_free=0);
	void showEvid (long evid);
	void setupEvUpdate();
	void setGlyphCallback (std::function<BQ_glyphMapeventsCallback> glyphCallback, BQ_ClientData client_data);

	BqMapevents vars;

protected:
	void drawViewportItem (QPainter *painter);
	void setVisibility () {};

private:
	void init(BQMap *map);
	void init(BQViewport *master, int make_navigator=1, int make_grids=1);
	void init(QWidget *parent, int make_navigator=1, int make_grids=1);
	void init2 (BQViewport *master);
	void timerCallback();
	void makeEventZoom ();
	void makeEventsGC ();
	void makeOriginsGC ();
	void hideEvents ();
	void showEvents ();
	void hideOrigins ();
	void showOrigins ();
	void makeArrivals (BUEVOrigin *origin);
	void hideArrivals ();
	void showArrivals ();
	void showEventPopup (QString text);
	int eventsShowMode (BQGlyph *glyph, BQEvent *event);
	int eventsDetectionsShowMode (BQGlyph *glyph, BQEvent *event);
	int originsShowMode (BQGlyph *glyph, BQEvent *event);
	int arrivalsShowMode (BQGlyph *glyph, BQEvent *event);
	QString makeEventText (BUEVEvent *event);
	QString makeOriginText (BUEVOrigin *origin);
	QString makeAssociationText (BUEVAssociation *association);
	QString makeDetectionText (BUEVDetection *detection);
	QString makeEventsDetectionText (BUEVDetection *detection);

	static int events_glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	static int events_detections_glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	static int origins_glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	static int arrivals_glyphs_callback (BQGlyph *glyph, BQEvent *event, BQ_ClientData clientData);
	static int coding_callback (BQGlyph *glyph, QString definitions_name, QString coding_name, double *value, BQ_ClientData client_data);
	static QList<BQ_XY> *makeGreatCirclePath (double lat1, double lon1, double lat2, double lon2, int n);
};

struct BQ_GraphicsServer;

typedef void (BQ_GraphicsServer_run_thread_callback) (BQ_GraphicsServer *server, int argc, char **argv);

#define BQ_GRAPHICSSERVER_TYPE_VIEWPORT		0
#define BQ_GRAPHICSSERVER_TYPE_AXES			1
#define BQ_GRAPHICSSERVER_TYPE_TEXT			2
#define BQ_GRAPHICSSERVER_TYPE_POLYLINE		3
#define BQ_GRAPHICSSERVER_TYPE_POLYPOINT	4
#define BQ_GRAPHICSSERVER_TYPE_MAP			5
#define BQ_GRAPHICSSERVER_TYPE_MAPEVENTS	6
#define BQ_GRAPHICSSERVER_TYPE_TEXTPOPUP	7
#define BQ_GRAPHICSSERVER_TYPE_TASKBAR		8

struct BQ_GraphicsServer_type {
	BQ_GraphicsServer_type () {
		type = 0;
		widget = NULL;
	}
	BQ_GraphicsServer_type (int t, void *w) {
		type = t;
		widget = w;
	}

	int type;
	void *widget;
};

#define BQ_GRAPHICSSERVER_STATE_INACTIVE	0
#define BQ_GRAPHICSSERVER_STATE_RUNNING		1
#define BQ_GRAPHICSSERVER_STATE_QUIT		2

class BQToplevel;

struct BQ_GraphicsServer {
	int argc;
	char **argv;
	QApplication *qapp;
	QString toplevel_name;
	Pmtfifo *send_mtfifo;
	Pmtfifo *receive_mtfifo;
	BQToplevel *toplevel;
	QHash<QString, BQ_GraphicsServer_type> handles;
	int handle_instance;
	BQ_GraphicsServer_run_thread_callback *run_thread_callback;
	pthread_t run_thread;
	pthread_t server_thread;
	int state;
};

class BQToplevel : public QWidget {
public:
	BQToplevel (BQ_GraphicsServer *server);
protected:
	void closeEvent (QCloseEvent *event);
	void paintEvent (QPaintEvent *event);
private:
	BQ_GraphicsServer *server;
	int first;
};

struct BQ_GraphicsServer_config_arg {
	BQ_GraphicsServer_config_arg (QString key, QString value) {
		this->key = key;
		this->value = value;
	}
	BQ_GraphicsServer_config_arg (char *key, char *value) {
		this->key = QString(key);
		this->value = QString(value);
	}

	QString key;
	QString value;
};

struct BQ_GraphicsServer_command {
	QString command;
	QString handle;
	QStringList args;
	QList<BQ_GraphicsServer_config_arg> config_args;
	QList<void *> extra_args;
};

#define BQ_GRAPHICSSERVER_RESPONSE_OK		0
#define BQ_GRAPHICSSERVER_RESPONSE_ERROR	1

struct BQ_GraphicsServer_response {
	BQ_GraphicsServer_response () {
		this->response = BQ_GRAPHICSSERVER_RESPONSE_OK;
	}
	BQ_GraphicsServer_response (int response, QString handle) {
		this->response = response;
		this->handle = handle;
	}

	int response;
	QString handle;
	QStringList return_params;
};

template <typename V> 
void BQ_showKeys(const char *pre, QHash<QString, V> *hash) {
	if (hash) return;
	QList<QString> keys = hash->keys();
	printf ("%s: keys: ", pre);
	foreach (QString key, keys) {
		printf ("%s ", key.toLocal8Bit().data());
	}
	printf ("\n");
}

extern "C" {
extern char *BQ_qevent_type_str (QEvent *event);
extern int BQ_Strcmp (QString str1, QString str2);
extern int BQ_ParseColor (QString value, BQ_Color &color);
extern void BQ_ClearColor (BQ_Color &color);
extern int BQ_ParseFont (QString value, BQ_Font &font);
extern void BQ_edp2latlon (double latr, double lonr, double delx, double dely, double *lat, double *lon);
extern void BQ_edp2latlonr (double lat, double lon, double delx, double dely, double *latr, double *lonr);
extern void BQ_globe2latlon (double latr, double lonr, double delx, double dely, double *lat, double *lon);
extern void BQ_globe2latlonr (double lat, double lon, double delx, double dely, double *latr, double *lonr);
extern void BQ_latlon2edp (double latr, double lonr, double lat, double lon, double *delx, double *dely);
extern void BQ_latlon2edp2 (double latr, double lonr, double lat, double lon, double *delx, double *dely, double *north);
extern int BQ_latlon2globe (double latr, double lonr, double lat, double lon, double *delx, double *dely);
extern int BQ_latlon2globe2 (double latr, double lonr, double lat, double lon, double *delx, double *dely, double *north, double *az=NULL, double *dip=NULL);
extern void BQ_delbox2llbox (double xlatc, double xlonc, double xdelmn, double xdelmx, double ydelmn, double ydelmx, int globe,
			double *xlatmn, double *xlatmx, double *xlonmn, double *xlonmx);
extern int BQ_setup_line (QPainter *painter, BQ_Color *color, int linewidth, int linestyle, int joinstyle);
extern int BQ_setup_fill (QPainter *painter, BQ_Color *color);
extern int BQ_setup_text (QPainter *painter, BQ_Color *color, BQ_Font *font);
extern int BQ_get_metrics (BQ_Font *font, char *text, int *w, int *h);
extern int BQ_get_fontmetrics (BQ_Font *font, char *text, int *str_w, int *str_h, int *ascent, int *descent);
extern int BQ_get_nice_fvals (double start, double end, double width, double minbigwidth,
			  double *start_big, int *n_big, double *incr_big,
			  double *start_sml, int *n_sml, double *incr_sml);
extern int BQ_nice_number (double number, char *format, char *string);
extern void BQ_InstallMessageHandler();
extern void BQ_MessageOutput(QtMsgType type, const QMessageLogContext &context, const QString &msg);
extern int BQ_pfcompile (char *pfstring, Pf **pf);
extern char *BQ_make_pfstring (QString pfstring, QString input_key, QString output_key);
extern BQ_GraphicsServer *BQ_GraphicsServer_create (int argc, char **argv, char *toplevel_name, BQ_GraphicsServer_run_thread_callback *run_thread_callback);
extern void *BQ_GraphicsServer_get_item (BQ_GraphicsServer *server, QString handle, int *type=NULL); 
extern int BQ_GraphicsServer_send (BQ_GraphicsServer *server, BQ_GraphicsServer_response **response, QString command, QString handle, QStringList *args, ...);
extern int BQ_GraphicsServer_send2 (BQ_GraphicsServer *server, BQ_GraphicsServer_response **response, QString command, QString handle, QStringList *args, QStringList *config_args);
extern int BQ_GraphicsServer_send_extra (BQ_GraphicsServer *server, BQ_GraphicsServer_response **response, QString command, QString handle, QList<void *> extra_args);
extern void BQ_GraphicsServer_release (BQ_GraphicsServer *server);
extern BQ_GraphicsServer *BQ_GraphicsServer_get_global ();
}

#endif
