/*   Copyright (c) 2015 Boulder Real Time Technologies, Inc.
 *
 *   This software module is wholly owned by Boulder Real Time
 *   Technologies, Inc. Any use of this software module without
 *   express written permission from Boulder Real Time Technologies,
 *   Inc. is prohibited.
 */

#ifndef QDBPTR_H
#define QDBPTR_H

#include <iostream>
#include <memory>
#include <QObject>
#include <QVariant>

#include "stock.h"
#include "db.h"

#define QDBPTR_NULL_MODTIME 	-9999999999.999

class QDsConnection;
class QDsResponse;

enum QDbptrType			{ QDBINVALID, QDATABASE, QTABLE, QFIELD, QRECORD, QCELL };
enum QDbptrToStringMode		{ QDBPTR_TOSTRING_MULTILINE, QDBPTR_TOSTRING_ONELINE, QDBPTR_TOSTRING_COMPACT };

class QDbptr : public QObject
{
	Q_OBJECT

	public:
				QDbptr( void );
				QDbptr( QDsConnection* connection, Dbptr db );
				~QDbptr( void );

		QDsConnection*	getConnection( void );

		bool		isValid( void );

		std::shared_ptr<QDsResponse>	request( QString request, QVariantList args );
		std::shared_ptr<QDsResponse>	request( QString request );

		QString		toString( QDbptrToStringMode mode = QDBPTR_TOSTRING_MULTILINE );

		// Convenience methods wrapping standard requests:

		QDbptrType	getDbptrType( void );
		Dbptr		getDbptr( void );
		long		getIndexDatabase( void );
		long		getIndexTable( void );
		long		getIndexField( void );
		long		getIndexRecord( void );

		bool		cellIsNull( void );
		QString		getName( void );
		QString		getDatabaseFileName( void );
		QString		getDatabaseName( void );
		QString		getSchemaName( void );
		QString		getTableName( void );
		QString		getFieldName( void );
		QString		getFieldNull( void );
		QString		getSchemaDescription( void );	
		QString		getDatabaseDescription( void );	
		QString		getTableDescription( void );	
		QString		getFieldDescription( void );	
		QString		getSchemaDetail( void );	
		QString		getDatabaseDetail( void );	
		QString		getTableDetail( void );	
		QString		getFieldDetail( void );	
		QString		getFieldRange( void );	
		QString		getFieldFormat( void );	
		QString		getFieldUnits( void );	
		QString		getFieldBaseTable( void );	
		QString		getDatabaseFilename( void );	
		QString		getTableFilename( void );	
		QString		getTableDirname( void );	
		QString		getDatabaseAbspath( void );	
		QString		getDbPath( void );	
		QString		getTableCreation( void );	
		QString		getUniqueIdName( void );	
		QString		getTimedateName( void );	
		QString		getDbIdServer( void );	
		QString		getDbLocks( void );	

		int		getFieldType( void );
		long		getTableCount( void );
		long		getFieldCount( void );
		long		getRecordCount( void );
		int		columnApproxWidthChars( void );
		long		getTableSize( void );	
		long		getFieldSize( void );	
		long		getRecordSize( void );	
		long		getFieldIndex( void );	

		double		getTableModTime( void ); 

		bool		tableIsPresent( void );
		bool		tableIsView( void );
		bool		tableIsWritable( void );	
		bool		columnIsProbablyNull( void );
		bool		fieldIsPrimary( void );

		QList<QString>	getSchemaTables( void );
		QList<QString>	getBaseTablesPresent( void );
		QList<QString>	getBaseTablesAbsent( void );
		QList<QString>	getViewsPresent( void );
		QList<QString>	getViewTables( void );
		QList<QString>	findJoinTables( void );
		QList<QString>	getTableFields( void );
		QList<QString>	getNonNullTableFields( void );
		QList<QString>	getPrimaryKey( void );
		QList<QString>	getAlternateKey( void );
		QList<QString>	getForeignKeys( void );

		QMap< QString, QList< QList<QString> > >	findJoinTablesPresent( void );

		QHash< QString, QString > 	getLinkFieldTables( void );
		QHash< QString, double >  	getTableModTimes( void ); 

		QDbptr*		lookupTable( QString tableName );
		QDbptr*		lookupTable( int table );
		QDbptr*		lookupField( QString fieldName );
		QDbptr*		lookupField( int field );
		QDbptr*		lookupRecord( int record );
		QDbptr*		subset( QString expr, QString newViewName = QString() );
		QDbptr*		sort( QList<QString> fields, QString newViewName = QString(), bool reverse = false, bool unique = false );
		QDbptr*		sort( QString field, QString newViewName = QString(), bool reverse = false, bool unique = false );
		QDbptr*		group( QList<QString> fields, QString newViewName = QString() );
		QDbptr*		ungroup( QString newViewName = QString() );
		QDbptr*		joinTable( QString tableName, QString newViewName = QString() );
		QDbptr*		outerJoinTable( QString tableName, QString newViewName = QString() );
		QDbptr*		leftJoinTable( QString tableName, QString newViewName = QString() );
		QDbptr*		noJoinTable( QString tableName, QString newViewName = QString() );
		QDbptr*		thetaJoinTable( QString tableName, QString expr, QString newViewName = QString() );
		QDbptr*		severTable( QString tableName, QString newViewName = QString() );
		QDbptr*		separateTable( QString tableName, QString newViewName = QString() );

		QString		dbGet( bool scratch = false );

		QVariant	dbGetv( void );	// this one is just for an indiv. cell (field+record specified)
		QVariantHash	dbGetv( QList<QString> fields );

		long		dbFind( QString expr, bool backwards = false );
		long		dbAdd( QString row = QString() );
		long		dbAddchk( QString row = QString() );

		void		dbPutv( QString fieldName, QVariant value );	// SCAFFOLD need overloads for multiple values
		void		dbMark( void );
		void		dbCrunch( void );
		void		dbFree( void );

	public slots:
		
		void		invalidate( void );
		void		closeDatabase( void );


	signals:
		void		dataChanged( void );
		void		invalidated( void );

	private:

		QList<QString>	QVariantListToQStringList( QVariantList listIn );
		
		QDsConnection*	m_dsConnection;
		Dbptr		m_db;
};

#endif
